/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Tue Mar  5 19:02:11 MSK 2002
    copyright            : (C) 2002 by Alexei Soloviev, Alexei Stadnik
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "stdmodel.h"
#include "sansmodel.h"
#include "yumomodel.h"

#include "lsf.h"

#include "getopt.h"
#include "usage.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cctype>
#include <algorithm>
#include <vector>
#include <string>
#include <cstring>
using namespace std;

int main ( int argc, char * argv[] )
{
  if ( argc == 1 ) {
    cerr << PACKAGE << ": missing arguments" << endl;
    cerr << "Try \"" << PACKAGE << " --help\" for more information." << endl;
    return 1;
  }

  /** Data stuff */
  vector<double> x, y, dy, dx;
  double * x_min = 0;
  double * x_max = 0;

  /** Model stuff */
  Model * model = 0;
  bool do_rc_read = false;
  bool do_rc_write = false;

  /** Fitting stuff */
  string strategy = "MINIMIZE";
  bool robust = false;
  int verbose = -1;

  /** Output stuff */
  ostream * info = &cout;
  ofstream * data = 0;
  ostream * verb = &cerr;

  int c;

  while ( 1 ) { // reading command line arguments
    c = getopt_long ( argc, argv, 
		      short_options(), long_options, 
		      &option_index );
    if ( c == -1 ) break;

    switch ( c ) {
    case 0:
      if ( !strcmp ( long_options[option_index].name, "std" ) )
	model = model_ptr<STDModel> ( optarg );
      if ( !strcmp ( long_options[option_index].name, "sans" ) )
	model = model_ptr<SANSModel> ( optarg );
      if ( !strcmp ( long_options[option_index].name, "yumo" ) )
	model = model_ptr<YuMOModel> ( optarg );
      if ( model == 0 ) {
      	*verb << PACKAGE
	     << ": see \"--" << long_options[option_index].name
	     << " help\" for list of models" << endl;
      	return 1;
      }
      break;

    case 'r':
      robust = true;
      break;

    case 'v':
      ++verbose;
      //if ( optarg ) vrbout = new ofstream ( optarg, ios::trunc|ios::app );
      if ( optarg ) verb = new ofstream ( optarg );
      break;

    case 's':
      strategy = optarg;
      break;

    case 'R':
      {
	string arg ( optarg );
       	string::iterator p = find ( arg.begin(), arg.end(), ':' );
        if ( p == arg.end() ) {
          *verb << PACKAGE << ": wrong range argument" << endl;
	  return 1;
        }
	string x1str ( arg.begin(), p );
	string x2str ( ++p, arg.end() );
        if ( !x1str.empty() ) x_min = new double ( atof ( x1str.c_str() ) );
        if ( !x2str.empty() ) x_max = new double ( atof ( x2str.c_str() ) );
        if ( x_min && x_max && *x_min >= *x_max ) {
	  *verb << PACKAGE << ": can't fit with an empty range!" << endl;
	  return 1;
        }
      }
      break;

    case 'o':
      //data = new ofstream ( optarg, ios::trunc|ios::app );
      data = new ofstream ( optarg );
      break;

    case 'H':
      //info = new ofstream ( optarg, ios::trunc|ios::app );
      info = new ofstream ( optarg );
      break;

    case '>':
      do_rc_read = true;
      break;

    case '<':
      do_rc_write = true;
      break;

    case 'V':
      cout << PACKAGE << " version " << VERSION << endl;
      return 0;
      break;

    case 'h':
      usage();
      return 0;
      break;

    default:
      return 1;
      break;
    }
  }

  ifstream in;
  if (  optind < argc ) {
    in.open ( argv[optind] );
    if ( !in ) {
      *verb << PACKAGE << ": cannot open input file \""
	   << argv[optind] << "\"" << endl;
      return 1;
    }
    if ( ++optind < argc ) {
      *verb << PACKAGE << ": wrong arguments: ";
      while ( optind < argc ) *verb << argv[optind++] << " ";
      *verb << endl;
      return 1;
    }
  } else {
    *verb << PACKAGE << ": no input file specified"<< endl;
    return 1;
  }

  if ( model == 0 ) {
    *verb << PACKAGE << ": no model specified"<< endl;
    return 1;
  }

  double ax, ay, ady, adx;
  char buff[1024];
  while ( in.getline ( buff, 1024, '\n' ) ) {
    istringstream line ( buff );
    if ( !( line >> ax >> ay >> ady ) ) {
    	*verb << PACKAGE << ": wrong input file format"<< endl;
    	return 1;
    }
    if ( x_min && ax < *x_min ) continue;
    if ( x_max && ax > *x_max ) continue;
    x.push_back(ax);
    y.push_back(ay);
    dy.push_back(ady);
    if ( line >> adx ) dx.push_back(adx);
  }
  if ( x.empty() ) {
    *verb << PACKAGE << ": all points undefined!"<< endl;
    return 1;
  }

  if ( YuMOModel *y = dynamic_cast<YuMOModel*>(model) ) {
    if ( !y->initResolution ( &*x.begin(), &*dx.begin(), dx.size() ) ) {
    	*verb << PACKAGE << ": input file containes no resolution data"<< endl;
    }
  }

  ChiSquare chi_square;
  chi_square.model ( model );
  chi_square.data ( &*x.begin(), &*y.begin(), &*dy.begin(), x.size() );

  if ( !chi_square.strategy ( strategy ) ) {
    *verb << PACKAGE << ": wrong minimizer" << endl;
    return 1;
  }

  if ( do_rc_read ) model->rc_read();
  chi_square.minimize ( robust, verbose );
  if ( do_rc_write ) model->rc_write();

  *info << chi_square;

  if ( data && data->is_open() ) {
    data->setf(ios::scientific,ios::floatfield);
    data->setf(ios::right,ios::adjustfield);
    data->precision(4);

    for ( vector<double>::const_iterator p = x.begin(); p != x.end(); ++p ) {
      *data << setw(12) << *p
	    << setw(12) << (*model)(*p) << endl;
    }
  }

  return 0;
}
