/* Bessel Functions of Integer Order */

/*
Sample page from NUMERICAL RECIPES IN C: THE ART OF SCIENTIFIC COMPUTING (ISBN 0-521-43108-5)
Copyright (C) 1988-1992 by Cambridge University Press.
Programs Copyright (C) 1988-1992 by Numerical Recipes Software. 
Permission is granted for internet users to make one paper copy for their own personal use. 
Further reproduction, or any copying of machine-readable files (including this one) to any server
computer, is strictly prohibited. To order Numerical Recipes books,
diskettes, or CDROMsvisit website http://www.nr.com or call 1-800-872-7423 (North America only),
or send email to trade@cup.cam.ac.uk (outside North America). 
*/

#include <math.h>
#include <stdio.h>

#include "bessel.h"

float bessj0(float x)  /** Returns the Bessel function J0(x) for any real x. */
{
  float ax,z;
  double xx,y,ans,ans1,ans2;  /** Accumulate polynomials in double precision. */

  if ((ax=fabs(x)) < 8.0) {  /** Direct rational function fit. */
    y=x*x;
    ans1=57568490574.0+y*(-13362590354.0
			  +y*(651619640.7+y*(-11214424.18+y*(77392.33017
							     +y*(-184.9052456)))));
    ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
					  +y*(59272.64853+y*(267.8532712+y*1.0))));
    ans=ans1/ans2;
  } else {  /** Fitting function (6.5.9). */
    z=8.0/ax;
    y=z*z;
    xx=ax-0.785398164; ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
						       +y*(-0.2073370639e-5
							   +y*0.2093887211e-6)));
    ans2 = -0.1562499995e-1+y*(0.1430488765e-3
			       +y*(-0.6911147651e-5+y*(0.7621095161e-6
						       -y*0.934935152e-7)));
    ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2); 
  }
  return ans;
}

float bessy0(float x)  /** Returns the Bessel function Y0(x) for positive x. */
{
  float bessj0(float x); 
  float z; 
  double xx,y,ans,ans1,ans2;  /** Accumulate polynomials in double precision. */

  if (x < 8.0) {  /** Rational function approximation of (6.5.8). */
    y=x*x;
    ans1 = -2957821389.0+y*(7062834065.0+y*(-512359803.6
					    +y*(10879881.29+y*(-86327.92757
							       +y*228.4622733))));
    ans2=40076544269.0+y*(745249964.8+y*(7189466.438
					 +y*(47447.26470+y*(226.1030244+y*1.0))));
    ans=(ans1/ans2)+0.636619772*bessj0(x)*log(x);
  } else {  /** Fitting function (6.5.10). */
    z=8.0/x;
    y=z*z;
    xx=x-0.785398164;
    ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
				    +y*(-0.2073370639e-5+y*0.2093887211e-6)));
    ans2 = -0.1562499995e-1+y*(0.1430488765e-3
			       +y*(-0.6911147651e-5+y*(0.7621095161e-6
						       +y*(-0.934945152e-7))));
    ans=sqrt(0.636619772/x)*(sin(xx)*ans1+z*cos(xx)*ans2); 
  }
  return ans;
}

float bessj1(float x)  /** Returns the Bessel function J1(x) for any real x. */
{
  float ax,z;
  double xx,y,ans,ans1,ans2;  /** Accumulate polynomials in double precision. */

  if ((ax=fabs(x)) < 8.0) {  /** Direct rational approximation. */
    y=x*x;
    ans1=x*(72362614232.0+y*(-7895059235.0+y*(242396853.1
					      +y*(-2972611.439+y*(15704.48260
								  +y*(-30.16036606))))));
    ans2=144725228442.0+y*(2300535178.0+y*(18583304.74
					   +y*(99447.43394+y*(376.9991397+y*1.0))));
    ans=ans1/ans2; 
  } else {  /** Fitting function (6.5.9). */
    z=8.0/ax;
    y=z*z;
    xx=ax-2.356194491;
    ans1=1.0+y*(0.183105e-2+y*(-0.3516396496e-4
			       +y*(0.2457520174e-5+y*(-0.240337019e-6))));
    ans2=0.04687499995+y*(-0.2002690873e-3
			  +y*(0.8449199096e-5+y*(-0.88228987e-6 +y*0.105787412e-6)));
    ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
    if (x < 0.0) ans = -ans; 
  }
  return ans;
}

float bessy1(float x)  /** Returns the Bessel function Y1(x) for positive x. */
{
  float bessj1(float x);
  float z;
  double xx,y,ans,ans1,ans2;  /** Accumulate polynomials in double precision. */

  if (x < 8.0) {  /** Rational function approximation of (6.5.8). */
    y=x*x;
    ans1=x*(-0.4900604943e13+y*(0.1275274390e13
				+y*(-0.5153438139e11+y*(0.7349264551e9
							+y*(-0.4237922726e7
							    +y*0.8511937935e4)))));
    ans2=0.2499580570e14+y*(0.4244419664e12
			    +y*(0.3733650367e10+y*(0.2245904002e8
						   +y*(0.1020426050e6
						       +y*(0.3549632885e3+y)))));
    ans=(ans1/ans2)+0.636619772*(bessj1(x)*log(x)-1.0/x);
  } else {  /** Fitting function (6.5.10). */
    z=8.0/x;
    y=z*z;
    xx=x-2.356194491;
    ans1=1.0+y*(0.183105e-2+y*(-0.3516396496e-4
			       +y*(0.2457520174e-5+y*(-0.240337019e-6))));
    ans2=0.04687499995+y*(-0.2002690873e-3
			  +y*(0.8449199096e-5+y*(-0.88228987e-6 +y*0.105787412e-6)));
    ans=sqrt(0.636619772/x)*(sin(xx)*ans1+z*cos(xx)*ans2);
  }
  return ans;
}

float bessy(int n, float x)  /** Returns the Bessel function Yn(x) for positive x and n >= 2. */
{
  float bessy0(float x);
  float bessy1(float x);
  int j;
  float by,bym,byp,tox;

  if (n < 2) {
    fprintf(stderr,"Index n less than 2 in bessy\n");
    return 0;
  }
  tox=2.0/x;
  by=bessy1(x);  /** Starting values for the recurrence. */
  bym=bessy0(x); 
  for (j=1;j<n;j++) {  /** Recurrence (6.5.7). */
    byp=j*tox*by-bym;
    bym=by;
    by=byp;
  }
  return by;
}

#define ACC 40.0  /** Make larger to increase accuracy. */
#define BIGNO 1.0e10
#define BIGNI 1.0e-10

float bessj(int n, float x)  /** Returns the Bessel function Jn(x) for any real x and n >= 2. */
{
  float bessj0(float x);
  float bessj1(float x);
  int j,jsum,m;
  float ax,bj,bjm,bjp,sum,tox,ans;

  if (n < 2) 
    {
      fprintf(stderr,"Index n less than 2 in bessj\n");
      return 0;
    }

  ax=fabs(x);
  if (ax == 0.0)
    return 0.0;
  else if (ax > (float) n) {  /** Upwards recurrence from J0 and J1. */
    tox=2.0/ax;
    bjm=bessj0(ax);
    bj=bessj1(ax);
    for (j=1;j<n;j++) {
      bjp=j*tox*bj-bjm;
      bjm=bj;
      bj=bjp;
    }
    ans=bj;
  } else {  /** Downwards recurrence from an even m here computed. */
    tox=2.0/ax;
    m=2*((n+(int) sqrt(ACC*n))/2);
    jsum=0;  /** jsum will alternate between 0 and 1;
		 when it is 1, we accumulate in sum the even terms in (5.5.16). */
    bjp=ans=sum=0.0;
    bj=1.0;
    for (j=m;j>0;j--) {  /** The downward recurrence. */
      bjm=j*tox*bj-bjp;
      bjp=bj;
      bj=bjm;
      if (fabs(bj) > BIGNO) { /** Renormalize to prevent overflows. */
	bj *= BIGNI;
	bjp *= BIGNI;
	ans *= BIGNI;
	sum *= BIGNI;
      }
      if (jsum) sum += bj;  /** Accumulate the sum. */
      jsum=!jsum;  /** Change 0 to 1 or vice versa. */
      if (j == n) ans=bjp;  /** Save the unnormalized answer. */
    }
    sum=2.0*sum-bj;  /** Compute (5.5.16) */
    ans /= sum;  /** and use it to normalize the answer. */
  }
  return x < 0.0 && (n & 1) ? -ans : ans;
}

