/***************************************************************************
   integral.cpp  -  Implementation of interface to FORTRAN integrator DCUHRE

      The following types are integrable:
       - ordinary functions: double (*) ( const double * );
       - types with a method: double operator() ( const double * x ) const;
         (they should be inherited from class Integrand)

                             -------------------
    begin                : Sat Jan 11 2003
    copyright            : (C) 2003 by Alexei Soloviev
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "integral.h"

namespace DCUHRE {  /* Fortran modules & interface to them */

  /* Fortran Subroutine */
  extern "C" int dcuhre_( int *ndim, int *numfun,
			  double *a, double *b,
			  int *minpts, int *maxpts,
			  int (*funsub) (int*, double*, int*, double*),
			  double *epsabs, double *epsrel,
			  int *key, int *nw, int *restar,
			  double *result, double *abserr,
			  int *neval, int *ifail,
			  double *work );

  /* Function need to integrate */
  const Integrand * function;

  /* Subroutine called by dcuhre */
  inline
    int sub ( int *ndim, double *z, int *nfun, double *f )
    {
      *f = (*function) ( z );
      return 0;
    }

  /* Interface to dcuhre */
  double integral ( const Integrand & f,  /* function need to integrate */
		    const int ndim, /* dimension of integral (ndim=2,...,15) */
		    const double * a, /* the lower limits of integration */
		    const double * b, /* the upper limits of integration */
		    const double & err ) /* required precision */
    {
      /* Remember function to integrate */
      function = &f;

      /* Table of constant values */
      static const int nfun = 1;
      static const int zero = 0;
      static const int nw = 100000;
      static const int maxcls = 500000;
      static const double absreq = 0.;

      /* Local variables */
      static int ifail, neval;
      static double absest[nfun], finest[nfun], wrkstr[nw];

      /* Function Body */
      dcuhre_(const_cast<int*>(&ndim),
	      const_cast<int*>(&nfun),
	      const_cast<double*>(a),
	      const_cast<double*>(b),
	      const_cast<int*>(&zero),
	      const_cast<int*>(&maxcls),
	      sub,
	      const_cast<double*>(&absreq),
	      const_cast<double*>(&err),
	      const_cast<int*>(&zero),
	      const_cast<int*>(&nw),
	      const_cast<int*>(&zero),
	      finest, absest,
	      &neval, &ifail, wrkstr);

      return finest[0];
    }
}

