/***************************************************************************
             integral.h  -  interface to FORTRAN integrator DCUHRE

      The following types are integrable:
       - ordinary functions: double (*) ( const double * );
       - types with a method: double operator() ( const double * x ) const;
         (they should be inherited from class Integrand)

                             -------------------
    begin                : Sat Jan 11 2003
    copyright            : (C) 2003 by Alexei Soloviev
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef INTEGRAL_H
#define INTEGRAL_H

/* Function need to integrate */
class Integrand {
  typedef double (*Function) ( const double * );
  const Function * __function;
 public:
  Integrand ( ) : __function(0) {}
  Integrand ( const Function & f ) : __function(&f) {}
  virtual ~Integrand ( ) {}
  virtual double operator() ( const double * x ) const 
    {
      return (*__function)(x);
    }
};

/* 1 < Dimension < 16 */
inline
double integral ( const Integrand & f,  /* function need to integrate */
		  const int ndim,  /* dimension of integral (ndim=2,...,15) */
		  const double * a,  /* the lower limits of integration */
		  const double * b,  /* the upper limits of integration */
		  const double & err = 0.1 );  /* required precision */

#include <vector>
typedef std::vector<double> Limits;

/* 1 < Dimension < 16 */
inline
double integral ( const Integrand & f,  /* function need to integrate */
		  const Limits & a,  /* the lower limits of integration */
		  const Limits & b,  /* the upper limits of integration */
		  const double & err = 0.1 )  /* required precision */
{
  return integral ( f, a.size(), &*a.begin(), &*b.begin(), err );
}

/* Dimension = 1 */ 
inline
double integral ( const Integrand & f,  /* function need to integrate */
		  const double & a,  /* the lower limit of integration */
		  const double & b,  /* the upper limit of integration */
		  const double & err = 0.1 )  /* required precision */
{
  static double av[] = {0,0}, bv[] = {1,1};
  av[0] = a;
  bv[0] = b;
  return integral ( f, 2, av, bv, err );
}

/***************************************************************************
 *                                                                         *
 *   Implementation                                                        *
 *                                                                         *
 ***************************************************************************/

namespace DCUHRE {  /* Fortran modules & interface to them */

  /* Interface to dcuhre */
  double integral ( const Integrand & f,  /* function need to integrate */
		    const int ndim, /* dimension of integral (ndim=2,...,15) */
		    const double * a, /* the lower limits of integration */
		    const double * b, /* the upper limits of integration */
		    const double & err ); /* required precision */

}

/*  1 < Dimension < 16 */
inline
double integral ( const Integrand & f,  /* function need to integrate */
		  const int ndim,  /* dimension of integral (ndim=2,...,15) */
		  const double * a,  /* the lower limits of integration */
		  const double * b,  /* the upper limits of integration */
		  const double & err )  /* required precision */
{
  return DCUHRE::integral ( f, ndim, a, b, err );
}

#endif

