/***************************************************************************
                          sansmodel.h  -  description
                             -------------------
    begin                : Sat Mar 9 2002
    copyright            : (C) 2002 by Alexei Soloviev, Alexei Stadnik
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SANSMODEL_H
#define SANSMODEL_H

#include "model.h"
#include "integral.h"

#include <map>

/**SANS models
  *@author Alexei Soloviev, Alexei Stadnik
  */

class SANSModel : public Model, public Integrand {
  typedef double (SANSModel::*IF) (const double *) const;
  typedef double (SANSModel::*TF) (const double &) const;

 protected: // Protected constructor
  SANSModel( const std::vector<Parameter> & p, 
	     const IF & f1, const TF & f2, 
	     const std::string & n );
 public:
  virtual ~SANSModel();

 public:
  /** Theoretical function */
  virtual double operator() ( const double & x ) const;

 public: 
  /** To construct model over comand line parameters */
  static const Dictionary<SANSModel> & dictionary();

 public:
  /** Ball constructor */
  static const SANSModel ball();
  /** Ellipsoid-2 constructor */
  static const SANSModel ellipsoid2();
  /** Ellipsoid-3 constructor */
  static const SANSModel ellipsoid3();
  /** Cylinder constructor */
  static const SANSModel cylinder();
  /** Cylinder-2 constructor */
  static const SANSModel cylinder2();
  /** Parallelepiped constructor */
  static const SANSModel parallelepiped();
  /** Spherical shell constructor */
  static const SANSModel sphericalShell();
  /** Ellipsoid-2 shell constructor */
  static const SANSModel ellipsoid2Shell();
  /** Ellipsoid-3 shell constructor */
  static const SANSModel ellipsoid3Shell();

 private: // Private attributes
  /** Pointer to function to be integrated for current model */
  const IF fIntegFuncPtr;
  /** Pointer to theoretical function for current model */
  const TF fTheorFuncPtr;
  /** Shared variables */
  mutable double __Q;

 private: // Private methods
  /** Function to be integrated */
  double operator() ( const double * ) const;
  /** Ball function to be integrated and its theoretical function */
  double ball(const double *) const;
  double ball(const double &) const;
  /** Ellipsoid-2 function to be integrated and its theoretical function */
  double ellipsoid2(const double *) const;
  double ellipsoid2(const double &) const;
  /** Ellipsoid-3 function to be integrated and its theoretical function */
  double ellipsoid3(const double *) const;
  double ellipsoid3(const double &) const;
  /** Cylinder function to be integrated and its theoretical function */
  double cylinder(const double *) const;
  double cylinder(const double &) const;
  /** Cylinder-2 function to be integrated and its theoretical function */
  double cylinder2(const double *) const;
  double cylinder2(const double &) const;
  /** Parallelepiped function to be integrated and its theoretical function */
  double parallelepiped(const double *) const;
  double parallelepiped(const double &) const;
  /** Spherical shell function to be integrated and its theoretical function */
  double sphericalShell(const double *) const;
  double sphericalShell(const double &) const;
  /** Ellipsoid-2 shell function to be integrated and its theoretical function */
  double ellipsoid2Shell(const double *) const;
  double ellipsoid2Shell(const double &) const;
  /** Ellipsoid-3 shell function to be integrated and its theoretical function */
  double ellipsoid3Shell(const double *) const;
  double ellipsoid3Shell(const double &) const;
};

#endif
