/***************************************************************************
                          stdmodel.cpp  -  description
                             -------------------
    begin                : Fri Apr 19 2002
    copyright            : (C) 2002 by Alexei Soloviev, Alexei Stadnik
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stdmodel.h"

#include <cmath>
#include <cstdio>
#include <functional>
#include <algorithm>
#include <numeric>

STDModel::STDModel( const std::vector<Parameter> & p,
		      const TF & f, const std::string & n )
  : Model(p,n), fTheorFuncPtr(f)
{}

STDModel::~STDModel()
{}

/** Theoretical function */
double STDModel::operator() ( const double & x)  const {
  return ( this->*fTheorFuncPtr ) ( x );
}

/** To construct model over comand line parameters */
const Dictionary<STDModel> & STDModel::dictionary() {
  static Dictionary<STDModel> fDictionary;
  if ( fDictionary.empty() ) {
    fDictionary["g"] = &gaussian;
    fDictionary["e"] = &exponential;
    fDictionary["p0"] = &polynomial0;
    fDictionary["p1"] = &polynomial1;
    fDictionary["p2"] = &polynomial2;
    fDictionary["p3"] = &polynomial3;
    fDictionary["p4"] = &polynomial4;
    fDictionary["p5"] = &polynomial5;
    fDictionary["p6"] = &polynomial6;
    fDictionary["p7"] = &polynomial7;
    fDictionary["p8"] = &polynomial8;
  }
  return fDictionary;
}

/** Gaussian constructor */
const STDModel STDModel::gaussian () {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Mean", 0,  0.1 ) );
  parm.push_back ( Parameter( "RMS", 1, 0.1 ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0,  0.1 ) );
  return STDModel ( parm, &STDModel::gaussian, "Gaussian" );
}

/** Gaussian theoretical function */
double STDModel::gaussian ( const double & x ) const {
  double arg = ( x - fParm[0].value() ) / fParm[1].value();
  arg *= arg;
  arg /= -2;
  return exp ( arg )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Exponential constructor */
const STDModel STDModel::exponential () {
  std::vector<Parameter> parm;
  parm.push_back ( Parameter( "Slope", -1,  0.1 ) );
  parm.push_back ( Parameter( "Intercept", 0, 0.1 ) );
  parm.push_back ( Parameter( "Amplitude", 1, 0.1 ) );
  parm.push_back ( Parameter( "Background", 0,  0.1 ) );
  return STDModel ( parm, &STDModel::exponential, "Exponential" );
}

/** Exponential theoretical function */
double STDModel::exponential ( const double & x ) const {
  return exp ( fParm[0].value() * x + fParm[1].value() )
    * fParm[fParm.size()-2].value()
    + fParm[fParm.size()-1].value();
}

/** Polynomial constructors for users */
const STDModel STDModel::polynomial0 () { return polynomial ((unsigned int)0); }
const STDModel STDModel::polynomial1 () { return polynomial ((unsigned int)1); }
const STDModel STDModel::polynomial2 () { return polynomial ((unsigned int)2); }
const STDModel STDModel::polynomial3 () { return polynomial ((unsigned int)3); }
const STDModel STDModel::polynomial4 () { return polynomial ((unsigned int)4); }
const STDModel STDModel::polynomial5 () { return polynomial ((unsigned int)5); }
const STDModel STDModel::polynomial6 () { return polynomial ((unsigned int)6); }
const STDModel STDModel::polynomial7 () { return polynomial ((unsigned int)7); }
const STDModel STDModel::polynomial8 () { return polynomial ((unsigned int)8); }

/** Polynomial constructor */
const STDModel STDModel::polynomial ( const unsigned int order )
{
  vector<Parameter> parm;
  char buff[256];
  for ( unsigned int i = 0; i < order+1; ++i ) {
    sprintf ( buff, "%d", i );
    parm.push_back ( Parameter( (string("p")+=buff), 1, 0.1 ) );
  }
  sprintf ( buff, "%d", order );
  return STDModel ( parm, &STDModel::polynomial,
                    (string("Polynomial of the order ")+=buff) );
}

/** Polynomial theoretical function */
double STDModel::polynomial ( const double & x ) const {
  double arg = 1, res = 0;
  for ( unsigned int i = 0; i < fParm.size(); ++i ) {
    res += fParm[i].value() * arg;
    arg *= x;
  }
  return res;
}
